/*
 * Copyright (c) 2010-2011, The MiCode Open Source Community (www.micode.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.micode.notes.model;

import android.appwidget.AppWidgetManager;
import android.content.ContentUris;
import android.content.Context;
import android.database.Cursor;
import android.text.TextUtils;
import android.util.Log;

import net.micode.notes.data.Notes;
import net.micode.notes.data.Notes.CallNote;
import net.micode.notes.data.Notes.DataColumns;
import net.micode.notes.data.Notes.DataConstants;
import net.micode.notes.data.Notes.NoteColumns;
import net.micode.notes.data.Notes.TextNote;
import net.micode.notes.tool.ResourceParser.NoteBgResources;

/**
 * WorkingNote 类用于管理便签的创建、加载、保存等操作。
 * 它封装了便签的各种属性和操作，包括便签的内容、模式、背景颜色、提醒日期等。
 */
public class WorkingNote {
    // 便签对象，用于管理便签的各种数据
    private Note mNote;
    // 便签的 ID
    private long mNoteId;
    // 便签的内容
    private String mContent;
    // 便签的模式
    private int mMode;
    // 便签的提醒日期
    private long mAlertDate;
    // 便签的修改日期
    private long mModifiedDate;
    // 便签的背景颜色 ID
    private int mBgColorId;
    // 便签小部件的 ID
    private int mWidgetId;
    // 便签小部件的类型
    private int mWidgetType;
    // 便签所在文件夹的 ID
    private long mFolderId;
    // 上下文对象，用于访问系统资源和执行操作
    private Context mContext;
    // 日志标签，用于调试日志输出
    private static final String TAG = "WorkingNote";
    // 标记便签是否已被删除
    private boolean mIsDeleted;
    // 便签设置更改监听器，用于监听便签设置的变化
    private NoteSettingChangedListener mNoteSettingStatusListener;

    // 数据查询的投影列，用于从数据库中查询数据
    public static final String[] DATA_PROJECTION = new String[] {
            DataColumns.ID,
            DataColumns.CONTENT,
            DataColumns.MIME_TYPE,
            DataColumns.DATA1,
            DataColumns.DATA2,
            DataColumns.DATA3,
            DataColumns.DATA4,
    };

    // 便签查询的投影列，用于从数据库中查询便签信息
    public static final String[] NOTE_PROJECTION = new String[] {
            NoteColumns.PARENT_ID,
            NoteColumns.ALERTED_DATE,
            NoteColumns.BG_COLOR_ID,
            NoteColumns.WIDGET_ID,
            NoteColumns.WIDGET_TYPE,
            NoteColumns.MODIFIED_DATE
    };

    // 数据 ID 列的索引
    private static final int DATA_ID_COLUMN = 0;
    // 数据内容列的索引
    private static final int DATA_CONTENT_COLUMN = 1;
    // 数据 MIME 类型列的索引
    private static final int DATA_MIME_TYPE_COLUMN = 2;
    // 数据模式列的索引
    private static final int DATA_MODE_COLUMN = 3;
    // 便签父文件夹 ID 列的索引
    private static final int NOTE_PARENT_ID_COLUMN = 0;
    // 便签提醒日期列的索引
    private static final int NOTE_ALERTED_DATE_COLUMN = 1;
    // 便签背景颜色 ID 列的索引
    private static final int NOTE_BG_COLOR_ID_COLUMN = 2;
    // 便签小部件 ID 列的索引
    private static final int NOTE_WIDGET_ID_COLUMN = 3;
    // 便签小部件类型列的索引
    private static final int NOTE_WIDGET_TYPE_COLUMN = 4;
    // 便签修改日期列的索引
    private static final int NOTE_MODIFIED_DATE_COLUMN = 5;

    /**
     * 新建便签的构造函数。
     *
     * @param context  上下文对象
     * @param folderId 便签所在文件夹的 ID
     */
    private WorkingNote(Context context, long folderId) {
        mContext = context;
        // 初始化提醒日期为 0，表示没有提醒
        mAlertDate = 0;
        // 初始化修改日期为当前时间
        mModifiedDate = System.currentTimeMillis();
        mFolderId = folderId;
        mNote = new Note();
        // 初始化便签 ID 为 0，表示新便签还未分配 ID
        mNoteId = 0;
        // 标记便签未被删除
        mIsDeleted = false;
        // 初始化便签模式为 0
        mMode = 0;
        // 初始化小部件类型为无效类型
        mWidgetType = Notes.TYPE_WIDGET_INVALIDE;
    }

    /**
     * 加载现有便签的构造函数。
     *
     * @param context  上下文对象
     * @param noteId   便签的 ID
     * @param folderId 便签所在文件夹的 ID
     */
    private WorkingNote(Context context, long noteId, long folderId) {
        mContext = context;
        mNoteId = noteId;
        mFolderId = folderId;
        // 标记便签未被删除
        mIsDeleted = false;
        mNote = new Note();
        // 加载便签信息
        loadNote();
    }

    /**
     * 加载便签的基本信息，包括父文件夹 ID、背景颜色 ID、小部件 ID 等。
     */
    private void loadNote() {
        // 查询便签信息
        Cursor cursor = mContext.getContentResolver().query(
                ContentUris.withAppendedId(Notes.CONTENT_NOTE_URI, mNoteId), NOTE_PROJECTION, null,
                null, null);

        if (cursor != null) {
            if (cursor.moveToFirst()) {
                // 获取便签所在文件夹的 ID
                mFolderId = cursor.getLong(NOTE_PARENT_ID_COLUMN);
                // 获取便签的背景颜色 ID
                mBgColorId = cursor.getInt(NOTE_BG_COLOR_ID_COLUMN);
                // 获取便签小部件的 ID
                mWidgetId = cursor.getInt(NOTE_WIDGET_ID_COLUMN);
                // 获取便签小部件的类型
                mWidgetType = cursor.getInt(NOTE_WIDGET_TYPE_COLUMN);
                // 获取便签的提醒日期
                mAlertDate = cursor.getLong(NOTE_ALERTED_DATE_COLUMN);
                // 获取便签的修改日期
                mModifiedDate = cursor.getLong(NOTE_MODIFIED_DATE_COLUMN);
            }
            // 关闭游标
            cursor.close();
        } else {
            // 记录错误日志
            Log.e(TAG, "No note with id:" + mNoteId);
            // 抛出异常，表示找不到指定 ID 的便签
            throw new IllegalArgumentException("Unable to find note with id " + mNoteId);
        }
        // 加载便签的数据
        loadNoteData();
    }

    /**
     * 加载便签的数据，包括便签内容和模式等。
     */
    private void loadNoteData() {
        // 查询便签的数据
        Cursor cursor = mContext.getContentResolver().query(Notes.CONTENT_DATA_URI, DATA_PROJECTION,
                DataColumns.NOTE_ID + "=?", new String[] {
                    String.valueOf(mNoteId)
                }, null);

        if (cursor != null) {
            if (cursor.moveToFirst()) {
                do {
                    // 获取数据的 MIME 类型
                    String type = cursor.getString(DATA_MIME_TYPE_COLUMN);
                    if (DataConstants.NOTE.equals(type)) {
                        // 如果是普通便签类型，获取便签内容和模式
                        mContent = cursor.getString(DATA_CONTENT_COLUMN);
                        mMode = cursor.getInt(DATA_MODE_COLUMN);
                        // 设置便签的文本数据 ID
                        mNote.setTextDataId(cursor.getLong(DATA_ID_COLUMN));
                    } else if (DataConstants.CALL_NOTE.equals(type)) {
                        // 如果是通话便签类型，设置通话数据 ID
                        mNote.setCallDataId(cursor.getLong(DATA_ID_COLUMN));
                    } else {
                        // 记录日志，表示遇到错误的便签类型
                        Log.d(TAG, "Wrong note type with type:" + type);
                    }
                } while (cursor.moveToNext());
            }
            // 关闭游标
            cursor.close();
        } else {
            // 记录错误日志
            Log.e(TAG, "No data with id:" + mNoteId);
            // 抛出异常，表示找不到指定 ID 的便签数据
            throw new IllegalArgumentException("Unable to find note's data with id " + mNoteId);
        }
    }

    /**
     * 创建一个空的便签。
     *
     * @param context         上下文对象
     * @param folderId        便签所在文件夹的 ID
     * @param widgetId        便签小部件的 ID
     * @param widgetType      便签小部件的类型
     * @param defaultBgColorId 便签的默认背景颜色 ID
     * @return 新创建的便签对象
     */
    public static WorkingNote createEmptyNote(Context context, long folderId, int widgetId,
            int widgetType, int defaultBgColorId) {
        // 创建一个新的便签对象
        WorkingNote note = new WorkingNote(context, folderId);
        // 设置便签的背景颜色 ID
        note.setBgColorId(defaultBgColorId);
        // 设置便签小部件的 ID
        note.setWidgetId(widgetId);
        // 设置便签小部件的类型
        note.setWidgetType(widgetType);
        return note;
    }

    /**
     * 加载指定 ID 的便签。
     *
     * @param context 上下文对象
     * @param id      便签的 ID
     * @return 加载的便签对象
     */
    public static WorkingNote load(Context context, long id) {
        return new WorkingNote(context, id, 0);
    }

    /**
     * 保存便签。
     *
     * @return 如果保存成功返回 true，否则返回 false
     */
    public synchronized boolean saveNote() {
        if (isWorthSaving()) {
            if (!existInDatabase()) {
                // 如果便签不在数据库中，创建一个新的便签 ID
                if ((mNoteId = Note.getNewNoteId(mContext, mFolderId)) == 0) {
                    // 记录错误日志
                    Log.e(TAG, "Create new note fail with id:" + mNoteId);
                    return false;
                }
            }
            // 同步便签数据到数据库
            mNote.syncNote(mContext, mNoteId);

            /**
             * 如果存在该便签的小部件，更新小部件的内容
             */
            if (mWidgetId != AppWidgetManager.INVALID_APPWIDGET_ID
                    && mWidgetType != Notes.TYPE_WIDGET_INVALIDE
                    && mNoteSettingStatusListener != null) {
                // 通知监听器小部件内容发生变化
                mNoteSettingStatusListener.onWidgetChanged();
            }
            return true;
        } else {
            return false;
        }
    }

    /**
     * 判断便签是否存在于数据库中。
     *
     * @return 如果存在返回 true，否则返回 false
     */
    public boolean existInDatabase() {
        return mNoteId > 0;
    }

    /**
     * 判断便签是否值得保存。
     *
     * @return 如果值得保存返回 true，否则返回 false
     */
    private boolean isWorthSaving() {
        if (mIsDeleted || (!existInDatabase() && TextUtils.isEmpty(mContent))
                || (existInDatabase() && !mNote.isLocalModified())) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * 设置便签设置更改监听器。
     *
     * @param l 监听器对象
     */
    public void setOnSettingStatusChangedListener(NoteSettingChangedListener l) {
        mNoteSettingStatusListener = l;
    }

    /**
     * 设置便签的提醒日期。
     *
     * @param date 提醒日期
     * @param set  是否设置提醒
     */
    public void setAlertDate(long date, boolean set) {
        if (date != mAlertDate) {
            mAlertDate = date;
            // 设置便签的提醒日期
            mNote.setNoteValue(NoteColumns.ALERTED_DATE, String.valueOf(mAlertDate));
        }
        if (mNoteSettingStatusListener != null) {
            // 通知监听器提醒日期发生变化
            mNoteSettingStatusListener.onClockAlertChanged(date, set);
        }
    }

    /**
     * 标记便签是否已被删除。
     *
     * @param mark 如果为 true 表示标记为已删除，否则为未删除
     */
    public void markDeleted(boolean mark) {
        mIsDeleted = mark;
        if (mWidgetId != AppWidgetManager.INVALID_APPWIDGET_ID
                && mWidgetType != Notes.TYPE_WIDGET_INVALIDE && mNoteSettingStatusListener != null) {
            // 通知监听器小部件内容发生变化
            mNoteSettingStatusListener.onWidgetChanged();
        }
    }

    /**
     * 设置便签的背景颜色 ID。
     *
     * @param id 背景颜色 ID
     */
    public void setBgColorId(int id) {
        if (id != mBgColorId) {
            mBgColorId = id;
            if (mNoteSettingStatusListener != null) {
                // 通知监听器背景颜色发生变化
                mNoteSettingStatusListener.onBackgroundColorChanged();
            }
            // 设置便签的背景颜色 ID
            mNote.setNoteValue(NoteColumns.BG_COLOR_ID, String.valueOf(id));
        }
    }

    /**
     * 设置便签的检查列表模式。
     *
     * @param mode 检查列表模式
     */
    public void setCheckListMode(int mode) {
        if (mMode != mode) {
            if (mNoteSettingStatusListener != null) {
                // 通知监听器检查列表模式发生变化
                mNoteSettingStatusListener.onCheckListModeChanged(mMode, mode);
            }
            mMode = mode;
            // 设置便签的文本数据模式
            mNote.setTextData(TextNote.MODE, String.valueOf(mMode));
        }
    }

    /**
     * 设置便签小部件的类型。
     *
     * @param type 小部件类型
     */
    public void setWidgetType(int type) {
        if (type != mWidgetType) {
            mWidgetType = type;
            // 设置便签的小部件类型
            mNote.setNoteValue(NoteColumns.WIDGET_TYPE, String.valueOf(mWidgetType));
        }
    }

    /**
     * 设置便签小部件的 ID。
     *
     * @param id 小部件 ID
     */
    public void setWidgetId(int id) {
        if (id != mWidgetId) {
            mWidgetId = id;
            // 设置便签的小部件 ID
            mNote.setNoteValue(NoteColumns.WIDGET_ID, String.valueOf(mWidgetId));
        }
    }

    /**
     * 设置便签的文本内容。
     *
     * @param text 文本内容
     */
    public void setWorkingText(String text) {
        if (!TextUtils.equals(mContent, text)) {
            mContent = text;
            // 设置便签的文本数据内容
            mNote.setTextData(DataColumns.CONTENT, mContent);
        }
    }

    /**
     * 将便签转换为通话便签。
     *
     * @param phoneNumber 电话号码
     * @param callDate    通话日期
     */
    public void convertToCallNote(String phoneNumber, long callDate) {
        // 设置通话便签的通话日期
        mNote.setCallData(CallNote.CALL_DATE, String.valueOf(callDate));
        // 设置通话便签的电话号码
        mNote.setCallData(CallNote.PHONE_NUMBER, phoneNumber);
        // 设置便签的父文件夹 ID 为通话记录文件夹的 ID
        mNote.setNoteValue(NoteColumns.PARENT_ID, String.valueOf(Notes.ID_CALL_RECORD_FOLDER));
    }

    /**
     * 判断便签是否设置了提醒。
     *
     * @return 如果设置了提醒返回 true，否则返回 false
     */
    public boolean hasClockAlert() {
        return (mAlertDate > 0 ? true : false);
    }

    public String getContent() {
        return mContent;
    }

    public long getAlertDate() {
        return mAlertDate;
    }

    public long getModifiedDate() {
        return mModifiedDate;
    }

    public int getBgColorResId() {
        return NoteBgResources.getNoteBgResource(mBgColorId);
    }

    public int getBgColorId() {
        return mBgColorId;
    }

    public int getTitleBgResId() {
        return NoteBgResources.getNoteTitleBgResource(mBgColorId);
    }

    public int getCheckListMode() {
        return mMode;
    }

    public long getNoteId() {
        return mNoteId;
    }

    public long getFolderId() {
        return mFolderId;
    }

    public int getWidgetId() {
        return mWidgetId;
    }

    public int getWidgetType() {
        return mWidgetType;
    }

    public interface NoteSettingChangedListener {
        /**
         * Called when the background color of current note has just changed
         */
        void onBackgroundColorChanged();

        /**
         * Called when user set clock
         */
        void onClockAlertChanged(long date, boolean set);

        /**
         * Call when user create note from widget
         */
        void onWidgetChanged();

        /**
         * Call when switch between check list mode and normal mode
         * @param oldMode is previous mode before change
         * @param newMode is new mode
         */
        void onCheckListModeChanged(int oldMode, int newMode);
    }
}
